#######################################################################
#                            ParameterDict                            #
#######################################################################


struct ParameterDict{K,V} <: AbstractDict{K,V}
  # (key, (value, default, isused))
  prms::OrderedDict{K,Tuple{V,V,Bool}}
end


function ParameterDict(p::Pair...)
  return ParameterDict(Dict(p))
end


function ParameterDict(d::AbstractDict{K,V}) where {K,V}
  prms = OrderedDict{K,Tuple{V,V,Bool}}()
  for (k,v) in pairs(d)
    prms[k] = (v, v, false)
  end
  sort!(prms)
  return ParameterDict{K,V}(prms)
end


function Base.setindex!(db::ParameterDict, new_v, key)
  vdu = db.prms[key]
  v, d, u = vdu
  db.prms[key] = (new_v, d, u)
  return db
end


function Base.getindex(db::ParameterDict, key)
  vdu = db.prms[key]
  v, d, u = vdu
  db.prms[key] = (v, d, true)
  return v
end


function setdefault!(db::ParameterDict, default, key)
  v, d, u = db.prms[key]
  db.prms[key] = (v, default, false)
  return db
end


function getdefault(db::ParameterDict, key)
  d = db.prms[key][2]
  return d
end


function reset(db::ParameterDict, key)
  d = getdefault(db, key)
  db.prms[key] = (d, d, false)
  return db
end


function reset(db::ParameterDict, default, key)
  db.prms[key] = (default, default, false)
  return db
end


function Base.get(db::ParameterDict, key, default)
  if haskey(db.prms, key)
    return db.prms[key][1]
  else
    return default
  end
end


Base.length(db::ParameterDict) = length(db.prms)


function Base.iterate(db::ParameterDict)
  it = iterate(db.prms)
  isnothing(it) && return
  pair, state = it
  k, v = pair
  return k => v[1], state
end


function Base.iterate(db::ParameterDict, state)
  it = iterate(db.prms, state)
  isnothing(it) && return
  pair, state = it
  k, v = pair
  return k => v[1], state
end


function Base.merge(db::ParameterDict, others::AbstractDict...)
  for os in others
    for (k,newval) in os
      setdefault!(db, newval, k)
    end
  end
  return newdb
end


function Base.pop!(db::ParameterDict, key, default)
  if !haskey(db.prms, key)
    return default
  else
    vdu = pop!(db.prms, key, default)
    return vdu[1]
  end
end


function isdefault(db::ParameterDict, key)
  v, d, _ = db.prms[key]
  return v == d
end


function isused(db::ParameterDict, key)
  return db.prms[key][3]
end


function Base.Dict(db::ParameterDict)
  return Dict(k => v for (k,v) in pairs(db))
end


function values_defaults_isused(db::ParameterDict{K,V}) where {K,V}
  vals = V[]
  defs = V[]
  used = Bool[]
  for (val,def,use) in values(db.prms)
    push!(vals, val)
    push!(defs, def)
    push!(used, use)
  end
  return vals, defs, used
end


function showparameters(db::ParameterDict; title=nothing, annotate=false)
  io = IOBuffer()
  showparameters(io, db; title, annotate)
  println(String(take!(io)))
  return
end


function showparameters(io::IO, db::ParameterDict; title=nothing, annotate=false)

  ks = collect(keys(db))
  vals, defs, uses = values_defaults_isused(db)
  isdef = [ v == d for (v,d) in zip(vals, defs) ]
  table = hcat(ks, vals)

  if annotate
    str_def    = [ d  ? "default" : "" for d in isdef ]
    str_unused = [ !u ? "not used" : "" for u in uses ]
    table = hcat(table, str_unused, str_def)
  end

  hl = Highlighter( (data, i, j) -> true,
      (h, data, i, j) -> begin
        d, u = isdef[i], uses[i]
        c = if !d && u
          crayon"green"
        elseif d && u
          crayon"blue"
        elseif d && !u
          crayon"dark_gray"
        else # !d && !u
          crayon"bold red"
        end
        c
      end)

  col_width = 39 # matches nice with TimerOutput in console
  pretty_table(io, table;
      body_hlines_format = Tuple('─' for _ = 1:4),
      alignment = annotate ? [:l, :l, :l, :l] : [:l, :l],
      highlighters = hl,
      title = isnothing(title) ? "" : title,
      title_alignment = :l,
      vlines = [],
      title_same_width_as_table = true,
      noheader = true,
      columns_width = annotate ? [ col_width, col_width, 10, 10 ] : col_width,
      linebreaks = true,
      autowrap = true,
      crop = :none
  )

  return
end


Base.show(io::IO, db::ParameterDict) = showparameters(io, db)
Base.show(io::IO, ::MIME"text/plain", db::ParameterDict) = showparameters(io, db)


#######################################################################
#                            ParameterDict                            #
#######################################################################


struct ParameterDB{K,V}
  dicts::Dict{K,ParameterDict{K,V}}
end


ParameterDB() = ParameterDB(Dict{String,ParameterDict{String,Any}}())


function ParameterDB(dicts::Pair...)
  return ParameterDB(Dict(dicts))
end


Base.getindex(db::ParameterDB, key) = db.dicts[key]
Base.setindex!(db::ParameterDB, dict, key) = db.dicts[key] = dict
Base.length(db::ParameterDB) = length(db.dicts)
Base.iterate(db::ParameterDB) = iterate(db.dicts)
Base.iterate(db::ParameterDB, state) = iterate(db.dicts, state)


function showparameters(io::IO, db::ParameterDB; annotate=false)
  for (sec, prms) in db.dicts
    showparameters(io, prms; title=sec, annotate)
  end
  return
end


function showparameters(db::ParameterDB; annotate=false)
  io = IOBuffer()
  showparameters(io, db; annotate)
  println(String(take!(io)))
  return
end


Base.show(io::IO, ::MIME"text/plain", db::ParameterDB) = showparameters(io, db)
Base.show(io::IO, db::ParameterDB) = showparameters(io, db)


function Base.Dict(db::ParameterDB)
  return Dict( k => Dict(v) for (k,v) in pairs(db.dicts) )
end


#######################################################################
#                             @parameters                             #
#######################################################################


"""
    @parameters name ex

Macro to define parameters and their default values for a symbol `name`.
If invoked with `MyProject` this generates the following methods for the `dg1d` project interface:
- `dg1d.parameters(::Val{:MyProject})`
- `dg1d.check_parameters(::Val{:MyProject}, prms)`

---

# Example

```julia
@parameters MyProject begin

  "some variable"
  x = 1.0
  @check x > 0.0

  "another variable"
  y = "String"

end
```
"""
macro parameters(name, expr)
  return esc(_generate_parameters(name, expr))
end


macro check(ex)
  return ex
end


function _generate_parameters(name, expr)

  if expr.head !== :block
    error("@parameters: Second argument must be a begin ... end block containing docstrings and parameter defintions!")
  end

  pars = expr.args
  pars = filter(p -> !(p isa LineNumberNode), pars)

  docstrings = String[]
  kvexprs = Expr[]
  checks = Vector{Expr}[]
  i_definition = 0
  for p in pars

    # if p isa Expr
    #   filter!(pp -> !(pp isa LineNumberNode), p.args)
    # end

    is_definition = (p isa Expr) && p.head === :macrocall &&
       (p.args[1] === Symbol("@doc") || p.args[1] isa GlobalRef) && length(p.args) == 4
    is_check = (p isa Expr) && p.head === :macrocall && p.args[1] === Symbol("@check") &&
      length(p.args) == 3

    is_valid_block = is_definition || is_check
    if !is_valid_block
      error("""
            @parameters: Invalid parameter definition.
            Expected either a block with a docstring, parameter name and default value
            or an argument check, e.g.
               \"this is a docstring for variable x\"
               x = 1.0
               @check x > 1.0
            Instead found
               $(string(p))
            """)
    end

    if is_definition
      # parse a parameter definition block

      i_definition += 1

      # extract docstring, mandatory
      d = p.args[3]
      if d isa Expr && d.head === :macrocall && d.args[1] === Symbol("@raw_str")
        d = d.args[3]
      end

      # extract variable name and default value, mandatory
      kv = p.args[4]
      if !(kv isa Expr)
        error("""
              @parameters: Invalid parameter definition. Missing default value for
                 $(string(kv))
              """)

      end
      if kv.head !== :(=)
        error("""
              @parameters: Invalid parameter definition. Expected pair of parameter name and default value, e.g.
                 x = 1.0
              Instead found
                 $(string(kv))
              """)
      end

      k, v = kv.args[1], kv.args[2]
      if !(k isa Symbol)
        error("""
              @parameters: Invalid parameter name: '$k'"
              """)
      end

      push!(docstrings, d)
      push!(kvexprs, kv)
      push!(checks, Expr[])
    end

    # parse a check, they are optional
    if is_check

      # we already verified above that p is a @check block
      # we also stripped any LineNumberNodes there
      ck = p.args[3]

      if i_definition == 0
        error("@parameters: Invalid definition. Found @check before any parameter definition: $ck")
      end

      # macro expansion of @check is delayed till check_parameters definition below
      push!(last(checks), ck)
    end

  end

  keys = [ kv.args[1] for kv in kvexprs ]
  for i = 1:length(keys)-1
    refk = keys[i]
    foreach(keys[i+1:end]) do k
      if refk === k
        error("""
              @parameters: Duplicated parameters found: '$k'
              """)
      end
    end
  end

  # generate the docs
  io = IOBuffer()
  println(io, "**[ ", name, " ]**")
  for (d, kv) in zip(docstrings, kvexprs)
    println(io, "- `", kv, "` : \n    ", replace(d, '\n' => "\n  "))
  end
  docs = String(take!(io))

  # compare key = value to "key" => value for Dict constructor
  pairs = [ Expr(:call, :(=>), string(kv.args[1]), kv.args[2]) for kv in kvexprs ]

  checkscode = Expr[]
  for cks in checks
    evalchecks = [ quote
                     result = try
                       $ck
                     catch
                       error("check_parameters for $($(string(name))): Failed to evaluate check '$($(string(ck)))'")
                     end
                     if !(result isa Bool)
                       error("check_parameters for $($(string(name))): Invalid check found '$($(string(ck)))'. Expected to return a Bool, but found $(typeof(result))")
                     end
                     if result == false
                       error("check_parameters for $($(string(name))): Check failed: '$($(string(ck)))'")
                     end
                   end
                   for ck in cks ]
    code = quote $(evalchecks...) end
    push!(checkscode, code)
  end
  unpack_prms = [ quote $k = prms["$($("$k"))"] end for k in keys ]

  code = quote
    @doc $docs dg1d.parameters(::Val{$(QuoteNode(name))}) = Dict($(pairs...))
    function dg1d.check_parameters(::Val{$(QuoteNode(name))}, prms)
      $(unpack_prms...)
      $(checkscode...)
    end
  end

  return code
end


#######################################################################
#                        Parameter definitions                        #
#######################################################################


@parameters Mesh begin

  """
  Number of cells
  - 1d: [ `kx` ]
  - 2d: [ `kx`, `ky` ]
  """
  k = 10
  @check k > 0
  @check length(k) in [ 1, 2 ]
  @check all(k .>= 0)

  """
  Physical domain extensions
  - 1d: [ `xmin`, `xmax` ]
  - 2d: [ `xmin`, `xmax`, `ymin`, `ymax` ]
  """
  range = [ -1.0, 1.0 ]
  @check length(range) in [ 2, 4 ]
  # TODO Need to introduce extra syntax (like -->) to only invoke some checks when needed, e.g.
  # @check length(range) == 2 --> range[2] - range[1] > 0.0
  # otherwise these tests fail always when run in 1d
  # @check length(range) >= 2 && range[2] - range[1] > 0.0
  # @check length(range) >= 4 && range[4] - range[3] > 0.0

  """
  Polynomial order in each grid
  - 1d: [ `nx` ]
  - 2d: [ `nx`, `ny` ]
  """
  n = 5
  @check length(n) in [ 1, 2 ]
  @check all(n .>= 0)

  # TODO Rename basis to quadrature
  """
  Nodal basis type used for the polynomial approximation. Available options
  - `lgl_lumped` ... mass-lumped Legendre-Gauss-Lobatto grid
  - `lgl`        ... Legendre-Gauss-Lobatto grid
  - `glgl`       ... Generalized Legendre-Gauss-Lobatto grid
  """
  basis = "lgl_lumped"
  @check basis in [ "lgl", "glgl", "lgl_lumped" ]

  """
  Cellwise approximation scheme of solution. Available options
  - `DG` ... Discontinuous Galerkin
  - `FV` ... Finite Volume (central scheme)
  """
  scheme = "DG"
  @check scheme in [ "DG", "FV" ]

  # TODO Rename this to basis
  """
  DG scheme used for the evolution. Available options
  - `nodal_lagrange` ... Nodal approximation using Lagrange polynomials collocated with
    the quadrature nodes provided by `basis`. This is the standard method discussed
    in Hesthaven, Warburton, 2007.
  - `modal_bernstein` ... A modal approximation using Bernstein polynomials.
    The coefficients correspond to the nodal values sampled on an equidistant grid,
    but the Bernstein polynomials are not interpolating.
    Hence, this Ansatz is also refered to as quasi-nodal.
    See Beisiegel, Behrens, Environ Earth Sci (2015) 74:7275–7284.
  """
  dg_kind = "nodal_lagrange"
  @check scheme in [ "nodal_lagrange", "modal_bernstein" ]

  """
  Periodicity in cartesian directions `x,y`
  - 1d: [ `x_dir` ]
  - 2d: [ `x_dir`, `y_dir` ]
  """
  periodic = [ true ]
  @check eltype(periodic) == Bool
  @check length(periodic) in [1,2,3]
  @check length(periodic)*2 == length(range)

  """
  Enable spectral filtering of solution after each full time step.
  """
  filter_enable = false
  @check filter_enable isa Bool

  """
  Parameters for exponential spectral filter.
  The parameters are `[ ηc, s/2, α ]` as defined in (5.16) in Hesthaven, Warburton 2007.
  The second parameter is halfed, because `s` needs to be even.
  """
  filter_parameters = [ 0.5, 16.0, 36.0 ]
  @check 0.0 ≤ filter_parameters[1] ≤ 1.0
  @check 0.0 ≤ filter_parameters[2]
  @check 0.0 ≤ filter_parameters[3]

end


@parameters Evolution begin

  "Final time"
  tend = 10.0
  @check tend > 0.0

  """
  Time stepping method. Available options
  - `midpt`         ... 2nd order explicit Runge Kutta method -- midpoint formula
  - `rk4`           ... 4th order explicit Runge Kutta method
  - `rk4_twothirds` ... 4th order explicit Runge Kutta method -- two thirds version
  - `rk4_ralston`   ... 4th order explicit Runge Kutta method -- Ralston's method, designed for minimal truncation error
  - `ssprk3`        ... 3rd order strong stability preserving Runge Kutta method
  - `lserk4`        ... 4th order low storage explicit Runge Kutta method
  - `rkf10`         ... 10th order Runge Kutta Feagin method
  """
  method = "lserk4"
  @check method in [ "midpt", "lserk4", "rk4", "rk4_twothirds", "rk_ralston",
                     "ssprk3", "rkf10" ]

  "Courant-Friedrichs-Lewy factor"
  cfl = 0.5
  @check cfl > 0.0

end


@parameters Output begin

  "Output data after `every_iteration` full time steps passed."
  every_iteration = 0
  @check every_iteration >= 0

  "Output data after `every_dt` simulation time passed."
  every_dt = 0.0
  @check every_dt >= 0

  """
  List of times at which data should be outputted.
  This will alter time stepping so that the output time is hit exactly.
  """
  aligned_ts = Float64[]
  @check all(t -> t > 0, aligned_ts)
  @check issorted(aligned_ts)

  """
  Interpolate and output data after `interpolate_every_iteration` full time steps.
  """
  interpolate_every_iteration = 0
  @check interpolate_every_iteration >= 0

  """
  Interpolate and output data after `interpolate_every_dt` simulation time.
  """
  interpolate_every_dt = 0.0
  @check interpolate_every_dt >= 0

  """
  List of times at which data should be interpolated and outputted.
  This will alter time stepping so that the output time is hit exactly.
  """
  interpolate_aligned_ts = Float64[]
  @check all(t -> t > 0, interpolate_aligned_ts)
  @check issorted(interpolate_aligned_ts)

  "Save timer log after `timer_every_dt_walltime` walltime (in seconds)."
  timer_every_dt_walltime = 1
  @check timer_every_dt_walltime >= 0

  "Print progress to console after `progress_every_dt` walltime (in seconds)."
  progress_every_dt_walltime = 0.2
  @check progress_every_dt_walltime >= 0

  """
  List of variables that should be outputted in 0d.
  Uses options `every_dt, every_iteration, aligned_ts`.
  """
  variables0d = String[]

  """
  List of variables that should be outputted in 1d.
  Uses options `every_dt, every_iteration, aligned_ts`.
  """
  variables1d = String[]

  """
  List of variables that should be outputted in 2d.
  Uses options `every_dt, every_iteration, aligned_ts`.
  """
  variables2d = String[]

  """
  List of variables that should be interpolated and outputted.
  """
  interpolate_variables = String[]

  """
  List of nodes onto which `interpolate_variables` should be interpolated and outputted.
  """
  interpolate_nodes = Float64[]

  "Basename for output"
  filename = ""
  @check length(filename) > 0

  """
  Data format for 0d output. Available options:
  - h5 ... raw hdf5 output, all fields in one file
  """
  format0d = "h5"
  @check format0d in [ "h5" ]

  """
  Data format for 1d output. Available options:
  - h5 ... raw hdf5 output, all fields in one file
  """
  format1d = "h5"
  @check format1d in [ "h5" ]

  """
  Data format for 2d output. Available options:
  - h5 ... raw hdf5 output, all fields in one file
  - xdmf+bin ... xdmf data description and binary data dump
  - xdmf+h5 ... xdmf data description and hdf5 data dump
  - xdmf ... same as xdmf+h5
  - vtk ... vtk xml format utilizing unstructured grids (but without subcell resolution)
  """
  format2d = "vtk"
  @check format2d in [ "h5", "xdmf", "xdmf+bin", "xdmf+h5", "vtk" ]

#######################################################################
#                       Substep output options                        #
#######################################################################

  "Output data from all substeps after `substep_every_iteration` full time steps passed."
  substep_every_iteration = 0
  @check substep_every_iteration >= 0

  "Output data from substeps after `every_dt` simulation time passed."
  substep_every_dt = 0.0
  @check substep_every_dt >= 0

  """
  List of variables that should be outputted from evolution substeps.
  Uses options `every_dt, every_iteration, aligned_ts`.
  """
  substep_variables = String[]

end


@parameters Log begin

  "Enable progress bar logging to stdout."
  progress_stdout = true

end
